//#############################################################################
//# Huge arrays and pointers for Microsoft Windows                            #
//#############################################################################
// Author: Paul Andre LeBlanc   CIS 71043,1643

#include "hugeclass.h"

//*****************************************************************************
//* class HugeAddress methods                                                 *
//*****************************************************************************

const LONG SEGSIZE = 65536L;
extern WORD _near _AHINCR;
extern WORD _near _AHSHIFT;
const WORD _ahincr = (WORD) &_AHINCR;
const WORD _ahshift = (WORD) &_AHSHIFT;

//-----------------------------------------------------------------------------
//- Huge pointer arithmetic                                                   -
//-----------------------------------------------------------------------------
// Assumes that any given segment contains x elements starting at offset 0x0000
// Address is adjusted so that individual elements do not cross segment
// boundaries.

LPVOID __HugeAdd(LPVOID address, LONG n, WORD size)
{
  WORD ofs;
  WORD sel = SELECTOROF(address);
  WORD elem_seg = SEGSIZE / size;
  while (n < 0)
  {
    n += elem_seg;
    sel -= _ahincr;
  }
  n += OFFSETOF(address) / size;
  sel += (n / elem_seg << _ahshift);
  ofs = n % elem_seg * size;
  return MAKELP(sel, ofs);
}

//-----------------------------------------------------------------------------
//- Allocate an array of elements                                             -
//-----------------------------------------------------------------------------

#ifdef	__DLL__
  #define ATYPE (GMEM_SHARE | GMEM_MOVEABLE) 	
#else
  #define ATYPE GPTR
#endif

LPVOID __HugeAlloc(LONG n, WORD size)
{
  WORD elems_seg = SEGSIZE / size;
  LONG data_size = n / elems_seg * SEGSIZE + n % elems_seg * size;
  return GlobalLock(GlobalAlloc(ATYPE, data_size));
}

//-----------------------------------------------------------------------------
//- Free an array of elements                                                 -
//-----------------------------------------------------------------------------

void __HugeFree(LPVOID address)
{
  HGLOBAL handle = GlobalHandle(SELECTOROF(address));
  GlobalUnlock(handle);
  GlobalFree(handle);
}
